package de.informatics4kids.hopfield;

public class Hopfield {

	/**
	 * Gre des Netzes beziehungsweise der zu erwartenden Bildern
	 */
	int size;

	/**
	 * Gewichtsmatrix W des Hopfield-Netzes
	 */
	MathMatrix weightMatrix;

	/**
	 * Dieser sog. Kontruktor wird bei der Erstellung des Netzes einmal
	 * automatisch aufgerufen, um das Netz zu konstruieren.
	 *
	 * @param size
	 *            - Die Gre des Netzes, also die sptere Anzahl der Pixel des
	 *            Bilds
	 */
	public Hopfield(int size) {
		this.size = size;
		weightMatrix = new MathMatrix(size, size);
	}

	/**
	 * Bringt dem Netz ein neues Bild bei. Die Bildanzahl sollte nicht mehr als
	 * 13% der Netzgre betragen!
	 *
	 * @param image
	 *            - Das zu erlernende Bild
	 */
	public void learn(MathVector image) {
		// auf richtige Gre prfen
		if (image.dimension() != size) {
			System.err.println("Falsche Bildgre!");
			return;
		}
		// Bildmatrix erstellen
		MathMatrix imageMatrix = createImageMatrix(image);
		// Diese Matrix zu der Gewichtsmatrix dazuaddieren
		weightMatrix.add(imageMatrix);
	}

	/**
	 * Lsst das Netz ein Bild vergessen.
	 * <p>
	 * Da die finale Gewichtsmatrix die Summe aller Einzelmatrizen der Bilder
	 * ist, kann durch Subtraktrion einer Bildmatrix ein Bild vergessen werden
	 * (da der Subtrahend den entsprechenden ursprnglichen Summanden aufhebt).
	 * <p>
	 * <b>Achtung:</b> Wenn ein Bild abgezogen wird, das zuvor noch nicht erlernt
	 * wurde, kann die Funktionalitt nicht weiterhin sichergestellt werden!
	 *
	 * @param image
	 *            - Das zu vergessende Bild
	 */
	public void forget(MathVector image) {
		// auf richtige Gre prfen
		if (image.dimension() != size) {
			System.err.println("Falsche Bildgre!");
			return;
		}
		// Bildmatrix erstellen
		MathMatrix imageMatrix = createImageMatrix(image);
		// Diese Matrix von der Gewichtsmatrix abziehen
		weightMatrix.subtract(imageMatrix);
	}

	/**
	 * Lsst das Netz alle Bilder vergessen, sodass es auf den Stand zum
	 * Zeitpunkt nach der Initialisierung zurckgesetzt ist.
	 */
	public void concussion() {
		// alle Werte wieder auf 0 setzen
		for (int i = 0; i < weightMatrix.rows; i++)
			for (int j = 0; j < weightMatrix.columns; j++)
				weightMatrix.matrix[i][j] = 0;
	}

	/**
	 * Erkennt ein unbekanntes Bild, falls das Originalbild bereits gelernt
	 * wurde, und nhert sich diesem an.
	 *
	 * @param unknownImage
	 *            - Ein unbekanntes Bild, dessen originale Form vorher bereits
	 *            gelernt wurde
	 * @return Ein Bild, das sich dem Original weiter angenhert hat, oder aber
	 *         direkt das Originalbild
	 */
	public MathVector recognize(MathVector unknownImage) {
		// Matrix-Vektor-Produkt der Gewichtsmatrix mit dem verrauschten (oder
		// anderweitig unkenntlicher gemachten) Bild errechnen
		MathVector result = weightMatrix.vectorProduct(unknownImage);
		// Ergebnis in die Eingabefunktion einsetzen
		result = entryFunction(result);
		return result;
	}

	/**
	 * Rekonstruiert das Originalbild vollstndig aus dem gegebenen Bild.
	 * Lediglich bei einer Oszillation wird die Rechnung abgebrochen. Verwendet
	 * fr die Berechnung die erkennen-Methode.
	 *
	 * @param unknownImage
	 *            - Ein unbekanntes Bild, dessen originale Form vorher bereits
	 *            gelernt wurde
	 * @return - Das rekonstruierte Originalbild oder bei einer Oszillation der
	 *         Oszillationspunkt (das Zwischenergebnis auf dem Weg zum
	 *         Originalbild sozusagen).
	 */
	public MathVector recognizeCompletely(MathVector unknownImage) {
		// Status, ob noch weitergerechnet werden muss
		boolean continueCalculation = true;

		// Wenn das Netz eine Lsung gefunden hat, verndert sich die Ausgabe im
		// Vergleich zur vorherigen nicht mehr. Um sie mit dieser vergleichen zu
		// knnen, muss die letze Ausgabe gespeichert werden.
		MathVector last = unknownImage;

		// Die nchste Ausgabe wird hier drin gespeichert.
		MathVector next = null;

		// Fr den Fall, dass das Netz zwischen zwei Zustnden oszilliert, muss
		// auch der vorletzte Stand gespeichert werden.
		MathVector oscillationTest = null;

		while (continueCalculation) {
			// Erkennen basierend auf dem letzen Ergebnis
			next = recognize(last);
			if (next.equals(last)) // keine Vernderung --> Lsung!
				continueCalculation = false;
			else if (next.equals(oscillationTest)) // Oszillation!
				continueCalculation = false;
			else { // noch keine Lsung --> weiterrechnen
				oscillationTest = last;
				last = next;
			}
		}
		return next;
	}

	/**
	 * Erstellt fr ein in vektorieller Form gegebenes Bild eine quadratische
	 * Bildmatrix mit einer der Bildgre gleichen Kantenlnge und somit
	 * Pixelanzahl*Pixelanzahl Elementen. Dabei sind die Diagonalelemente 0,
	 * whrend alle Elemente w_ij = w_ji sind. Jedes Element w_ij wird als
	 * Produkt aus der i-ten und der j-ten Komponente des Bildvektors gebildet.
	 *
	 * @param image
	 *            - Das gegebene Bild als Vektor
	 * @return Die Bildmatrix
	 */
	private MathMatrix createImageMatrix(MathVector image) {
		MathMatrix w = new MathMatrix(image.dimension(), image.dimension());
		for (int i = 0; i < w.rows; i++) {
			// Alle Diagonalelemente w_ii sind 0:
			w.matrix[i][i] = 0;
			for (int j = 0; j < i; j++) {
				// Der Wert w_ij ist das Produkt der Elemente i und j des
				// Bildvektors:
				int value_ij = image.tuple[i] * image.tuple[j];
				w.matrix[i][j] = value_ij;
				// Dieser Wert gilt auch fr w_ji:
				w.matrix[j][i] = value_ij;
			}
		}
		return w;
	}

	/**
	 * Erstellt einen neuen Vektor, indem jeder Wert des gegebenen Vektors in die
	 * Eingabefunkiton eingesetzt wird.
	 *
	 * @param v
	 *            - Ein Vektor, auf den die Eingabefunktion angewandt werden soll
	 * @return Ein mit der Eingabefunktion bearbeiteter Vektor
	 */
	private MathVector entryFunction(MathVector v) {
		MathVector result = new MathVector(v.dimension());
		for (int i = 0; i < v.dimension(); i++)
			result.tuple[i] = f(v.tuple[i]);
		return result;
	}

	/**
	 * Eingabefunktion fr einen Wert. Es wird -1 bei Eingaben unter 0
	 * zurckgegeben, sonst 1.
	 *
	 * @param x
	 *            - Eingabewert
	 * @return - Entweder -1 oder 1
	 */
	private int f(int x) {
		if (x < 0)
			return -1;
		else
			return 1;
	}

}
